#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.tests import common


class TestException(common.HeatTestCase):

    class SampleExceptionNoErorCode(exception.HeatException):
        msg_fmt = 'Test exception'

    class SampleException(SampleExceptionNoErorCode):
        error_code = 100

    def test_heat_exception_no_error_code(self):
        ex = TestException.SampleExceptionNoErorCode()
        self.assertEqual('Test exception',
                         ex.message)

    def test_heat_exception_with_error_code(self):
        ex = TestException.SampleException()
        self.assertEqual('HEAT-E100 Test exception',
                         ex.message)
