#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _


class EntityMatchNotFound(exception.HeatException):
    msg_fmt = _("No %(entity)s matching %(args)s.")

    def __init__(self, entity=None, args=None, **kwargs):
        super(EntityMatchNotFound, self).__init__(entity=entity, args=args,
                                                  **kwargs)


class EntityUniqueMatchNotFound(EntityMatchNotFound):
    msg_fmt = _("No %(entity)s unique match found for %(args)s.")


class InterfaceNotFound(exception.HeatException):
    msg_fmt = _("No network interface found for server %(id)s.")
