#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.clients import client_plugin


class DefaultClientPlugin(client_plugin.ClientPlugin):
    """A ClientPlugin that has no client.

    This is provided so that Resource can make use of the is_not_found() and
    is_conflict() methods even if the resource plugin has not specified a
    client plugin.
    """
    def _create(self, version=None):
        return super(DefaultClientPlugin, self)._create(version)
