#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.resources.openstack.senlin import res_base


class Receiver(res_base.BaseSenlinResource):
    """A resource that creates Senlin Receiver.

    Receiver is an abstract resource created at the senlin engine
    that can be used to hook the engine to some external event/alarm sources.
    """
    pass


def resource_mapping():
    return {
        'OS::Senlin::Receiver': Receiver
    }
